@@ -25,12 +25,16 @@
 # https://github.com/clowwindy/ssloop
 
 
+import os
+import socket
 import select
+import errno
+import logging
 from collections import defaultdict
 
 
 __all__ = ['EventLoop', 'POLL_NULL', 'POLL_IN', 'POLL_OUT', 'POLL_ERR',
-           'POLL_HUP', 'POLL_NVAL']
+           'POLL_HUP', 'POLL_NVAL', 'EVENT_NAMES']
 
 POLL_NULL = 0x00
 POLL_IN = 0x01
@@ -40,6 +44,16 @@ POLL_HUP = 0x10
 POLL_NVAL = 0x20
 
 
+EVENT_NAMES = {
+    POLL_NULL: 'POLL_NULL',
+    POLL_IN: 'POLL_IN',
+    POLL_OUT: 'POLL_OUT',
+    POLL_ERR: 'POLL_ERR',
+    POLL_HUP: 'POLL_HUP',
+    POLL_NVAL: 'POLL_NVAL',
+}
+
+
 class EpollLoop(object):
 
     def __init__(self):
@@ -142,18 +156,24 @@ class EventLoop(object):
     def __init__(self):
         if hasattr(select, 'epoll'):
             self._impl = EpollLoop()
+            model = 'epoll'
         elif hasattr(select, 'kqueue'):
             self._impl = KqueueLoop()
+            model = 'kqueue'
         elif hasattr(select, 'select'):
             self._impl = SelectLoop()
+            model = 'select'
         else:
             raise Exception('can not find any available functions in select '
                             'package')
         self._fd_to_f = {}
+        self._handlers = []
+        self.stopping = False
+        logging.debug('using event model: %s', model)
 
     def poll(self, timeout=None):
         events = self._impl.poll(timeout)
-        return ((self._fd_to_f[fd], event) for fd, event in events)
+        return [(self._fd_to_f[fd], fd, event) for fd, event in events]
 
     def add(self, f, mode):
         fd = f.fileno()
@@ -169,6 +189,25 @@ class EventLoop(object):
         fd = f.fileno()
         self._impl.modify_fd(fd, mode)
 
+    def add_handler(self, handler):
+        self._handlers.append(handler)
+
+    def run(self):
+        while not self.stopping:
+            try:
+                events = self.poll(1)
+            except (OSError, IOError) as e:
+                if errno_from_exception(e) == errno.EPIPE:
+                    # Happens when the client closes the connection
+                    continue
+                else:
+                    logging.error(e)
+                    continue
+            for handler in self._handlers:
+                # no exceptions should be raised by users
+                # TODO when there are a lot of handlers
+                handler(events)
+
 
 # from tornado
 def errno_from_exception(e):
@@ -187,3 +226,9 @@ def errno_from_exception(e):
         return e.args[0]
     else:
         return None
+
+
+# from tornado
+def get_sock_error(sock):
+    error_number = sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
+    return socket.error(error_number, os.strerror(error_number))
